/*
    Copyright (C) 2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: curve.h,v 1.9 2003/02/14 18:22:12 pbd Exp $
*/

#ifndef __ardour_curve_h__
#define __ardour_curve_h__

#include <sys/types.h>
#include <sigc++/signal_system.h>
#include <pbd/lockmonitor.h>
#include <pbd/undo.h>
#include <vector>
#include <pthread.h>

namespace ARDOUR {

class Curve : public SigC::Object
{
  public:
	Curve (double minval, double maxval, float defaultvalue);
	Curve (const Curve&);
	~Curve ();
	Curve& operator= (const Curve&);

	UndoAction get_memento ();
	SigC::Signal0<void> Changed;

	void get_vector (double x0, double x1, float *arg, size_t veclen);
	void add_control_point (double x, double y);
	void remove_control_point (size_t);
	void remove_control_points (size_t, size_t);
	void reset_control_point (size_t n, double xval, double yval);
	void reset_control_points (size_t, size_t, double);
	void truncate_end (double last_coordinate);
	void truncate_start (double overall_length);

	size_t n_control_points() const { return npoints; }
	
	bool closest_control_point_before (double xval, size_t& nth);
	bool closest_control_point_after (double xval, size_t& nth);
	bool control_points_adjacent (double xval, size_t& before, size_t& after,
				      double& before_x, double& after_x);

	bool nth (size_t n, double& xval, double &yval) {
		if (n < npoints) {
			xval = x[n];
			yval = y[n];
			return true;
		}
		return false;
	}

	void clear ();
	void solve ();

	template<class T> void apply_to_points (T& obj, void (T::*method)(size_t, double *, double *y)) {
		(obj.*method)(npoints, x, y);
	}

  private:
	double *x;
	double *y;
	float *second_derivative;
	size_t second_derivative_size;
	size_t npoints;
	double min_x;
	double max_x;
	double minval;
	double maxval;
	float  default_value;
	size_t size;
	PBD::Lock lock;
	bool _dirty;
	size_t last_lo;
	size_t last_hi;


	inline void spline_bound (size_t n, double *x, double val, size_t& lo, size_t& hi) {
		
		if (val >= x[last_lo] && val <= x[last_hi]) {
		    lo = last_lo;
		    hi = last_hi;
		    return;
		}

		/* do a binary search for the right interval: */

		size_t k_lo = 0; 
		size_t k_hi = n - 1;
		size_t k;

		while (k_hi - k_lo > 1)
		{
			k = (k_hi + k_lo) / 2;
			if (x[k] > val)
				k_hi = k;
			else
				k_lo = k;
		}

		lo = k_lo;
		hi = k_hi;
		last_lo = lo;
		last_hi = hi;
	}

	inline double spline_eval (double val)
	{
		double h, b, a;
		size_t lo, hi;
		spline_bound (npoints, x, val, lo, hi);
		h = x[hi] - x[lo];
		a = (x[hi] - val) / h;
		b = (val - x[lo]) / h;
		return a*y[lo] + b*y[hi] + ((a*a*a - a)*second_derivative[lo] + (b*b*b - b)*second_derivative[hi]) * (h*h)/6.0;
	}
	
	void _solve ();
	void _get_vector (double x0, double x1, float *arg, size_t veclen);
	void mark_dirty ();
	void copy_from_pointer (const Curve* copy);
};

}; /* namespace ARDOUR */

#endif /* __ardour_gain_h__ */
