/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: playlist.h,v 1.31 2003/02/18 17:59:22 pbd Exp $
*/

#ifndef __ardour_playlist_h__
#define __ardour_playlist_h__

#include <string>
#include <set>
#include <map>
#include <list>

#include <sys/stat.h>

#include <glib.h>
#include <sigc++/signal_system.h>
#include <pbd/atomic.h>
#include <pbd/undo.h>

#include <ardour/ardour.h>
#include <ardour/overlap.h>
#include <ardour/location.h>
#include <ardour/region.h>
#include <ardour/stateful.h>

namespace ARDOUR  {

class Session;

class Playlist : public Stateful, public SigC::Object {
  private:
	typedef set<Overlap*,Overlap::ComparePtr> Overlaps;
	
  public:
	typedef list<Region *> RegionList;

	struct State {
	    string operation;
	    RegionList regions;
	    list<UndoAction> region_states;
	
	    State (const string& why) : operation (why) {}
	};

	typedef vector<State*> StateMap;

	enum EditMode {
		Slide,
		Splice,
	};

	Playlist (Session&, const XMLNode&, bool hidden = false);
	Playlist (Session&, string name, bool hidden = false);
	Playlist (const Playlist&, string name, bool hidden = false);
	Playlist (const Playlist&, jack_nframes_t start, jack_nframes_t cnt, string name, bool hidden = false);

	Playlist& ref() { _refcnt++; return *this; }
	void unref() { if (_refcnt) _refcnt--; if (_refcnt == 0) delete this; }

	const string& name() const { return _name; }
	void set_name (const string& str);

	bool hidden() const { return _hidden; }
	
	EditMode get_edit_mode() const { return _edit_mode; }
	void set_edit_mode (EditMode);

	/* Editing operations */

	void add_region (const Region&, jack_nframes_t position, float times = 1, bool with_save = true);
	void remove_region (Region *);
	void replace_region (Region& old, Region& newr, jack_nframes_t pos);
	void split_region (Region&, jack_nframes_t position);
	void partition (jack_nframes_t start, jack_nframes_t end, bool just_top_level);
	void duplicate (Region&, jack_nframes_t position, float times);

	Playlist* cut  (list<AudioRange>&, bool result_is_hidden = true);
	Playlist* copy (list<AudioRange>&, bool result_is_hidden = true);
	int       paste (Playlist&, jack_nframes_t position, float times);

	void remove_matching_regions (Region&);

	layer_t top_layer() const;

	void clear (bool with_delete = false, bool with_save = true);

	jack_nframes_t get_maximum_extent () const;

	jack_nframes_t read       (Sample *dst, Sample *mixdown, float *gain_buffer, jack_nframes_t start, jack_nframes_t cnt, unsigned int chan_n=0);
 	int       read_peaks (PeakData *, jack_nframes_t npeaks, jack_nframes_t start, jack_nframes_t cnt, unsigned int chan_n=0);

	int set_loop_location (Location *loc);

	RegionList *regions_at (jack_nframes_t frame);
	Region     *top_region_at (jack_nframes_t frame);

	template<class T> void foreach_overlap (T *t, void (T::*func)(Overlap *)) {
		RegionLock rlock (this, false);
		for (Overlaps::iterator i = _overlaps.begin(); i != _overlaps.end(); i++) {	
			(t->*func) (*i);
		}
	}

	template<class T> void foreach_region (T *t, void (T::*func)(Region *, void *), void *arg) {
		RegionLock rlock (this, false);
		for (RegionList::iterator i = regions.begin(); i != regions.end(); i++) {	
			(t->*func) ((*i), arg);
		}
	}

	template<class T> void foreach_region (T *t, void (T::*func)(Region *)) {
		RegionLock rlock (this, false);
		for (RegionList::const_iterator i = regions.begin(); i != regions.end(); i++) {
			(t->*func) (*i);
		}
	}

	XMLNode& get_state ();
	int set_state (const XMLNode&);
	XMLNode& get_template ();

	SigC::Signal1<void,Region *> RegionAdded;
	SigC::Signal1<void,Region *> RegionRemoved;

	SigC::Signal1<void,Overlap *> NewRegionOverlap; 

	SigC::Signal0<void>            Modified;
	SigC::Signal0<void>            NameChanged;
	SigC::Signal0<void>            LengthChanged;
	SigC::Signal0<void>            LayeringChanged;
	SigC::Signal1<void,Location *> LoopSet;
	SigC::Signal1<void,Playlist *> GoingAway;
	SigC::Signal0<void>            StateChanged;
	SigC::Signal0<void>            StatePushed;

	static SigC::Signal1<void,Playlist*> PlaylistCreated;

	UndoAction get_memento() const;
	void       set_state (state_id_t);

	template<class T> void apply_to_history (T& obj, void (T::*method)(const StateMap&, state_id_t)) {
		RegionLock rlock (this);
		(obj.*method) (states, _current_state_id);
	}

	unsigned long refcnt() const { return _refcnt; }

	static string bump_name (string old_name);

	void freeze () { delay_notifications(); }
	void thaw ()   { release_notifications(); }

	int vacate_layer (layer_t, Region& r, int dir);

	/* this is ONLY for debugging */

	void dump ();

  protected:
	friend class Session;
	~Playlist ();  /* members of the public use unref() */

  private:
	
	struct RegionLock {
	    RegionLock (Playlist *pl, bool do_block_notify = true) : playlist (pl), block_notify (do_block_notify) {
		    playlist->region_lock.lock();
		    if (block_notify) {
			    playlist->delay_notifications();
		    }
	    }
	    ~RegionLock() { 
		    playlist->region_lock.unlock();
		    if (block_notify) {
			    playlist->release_notifications ();
		    }
	    }
	    Playlist *playlist;
	    bool block_notify;
	};

	friend class RegionLock;

	RegionList       regions;
	Location        *loop_location;
	string          _name;
	mutable PBD::NonBlockingLock region_lock;
	Session&         session;
	atomic_t         block_notifications;
	RegionList       pending_removals;
	RegionList       pending_adds;
	bool             pending_modified;
	bool             pending_length;
	bool             pending_layering;
	bool             pending_state_changed;
	bool             in_set_state;
	bool            _hidden;
	bool            _splicing;
	unsigned long   _refcnt;
	EditMode        _edit_mode;
	Overlaps        _overlaps;
	state_id_t      _current_state_id;
	StateMap         states;
	bool             in_flush;
	unsigned long   subcnt;

	void init (bool hide);

	void save_state (const string& why);
	void store_state (State *) const;
	void restore_state (State *);

	/* prevent the compiler from ever generating these */

	Playlist (const Playlist&);
	Playlist (Playlist&);

	void delay_notifications ();
	void release_notifications ();
	void flush_notifications ();

	void notify_region_removed (Region *);
	void notify_region_added (Region *);
	void notify_length_changed ();
	void notify_layering_changed ();
	void notify_modified ();
	void notify_state_changed ();

	void mark_session_dirty();

	void region_deleted (Region *);
	void region_bounds_changed (Region::BoundsChangeInfo, void *, Region *);
	void region_layer_changed ();
	void region_changed ();

	void sort_regions ();

	void possibly_splice ();
	void splice ();

	void add_region_internal (Region *, jack_nframes_t position, bool delay_sort = false);
	int remove_region_internal (Region *, bool delay_sort = false);
	RegionList *find_regions_at (jack_nframes_t frame);
	void copy_regions (RegionList&) const;
	void partition_internal (jack_nframes_t start, jack_nframes_t end, bool cutting, RegionList& thawlist);

	void check_overlaps (Region *region);
	void rebuild_overlaps ();
	void remove_overlaps (Region *region);

	XMLNode& state (bool);

	jack_nframes_t _get_maximum_extent() const;

	Playlist* cut_copy (Playlist* (Playlist::*pmf)(jack_nframes_t, jack_nframes_t, bool), 
			    list<AudioRange>& ranges, bool result_is_hidden);
	Playlist *cut (jack_nframes_t start, jack_nframes_t cnt, bool result_is_hidden);
	Playlist *copy (jack_nframes_t start, jack_nframes_t cnt, bool result_is_hidden);
};

} /* namespace ARDOUR */

#endif	/* __ardour_playlist_h__ */


