/*
    Copyright (C) 2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: redirect.h,v 1.21 2003/02/14 03:33:17 pbd Exp $
*/

#ifndef __ardour_redirect_h__
#define __ardour_redirect_h__

#include <vector>
#include <set>
#include <glib.h>
#include <sigc++/signal_system.h>
#include <string.h>

#include <pbd/lockmonitor.h>
#include <pbd/undo.h>

#include <ardour/ardour.h>
#include <ardour/io.h>
#include <ardour/automation_event.h>

class XMLNode;

namespace LADSPA {
	class Plugin;
}

namespace ARDOUR {

class Session;

class Redirect : public IO
{
  public:
	static const string state_node_name;

	enum Placement {
		PreFader,
		PostFader
	};

	Redirect (Session&, const string& name, Placement,
		  int input_min = -1, int input_max = -1, int output_min = -1, int output_max = -1);
	Redirect (const Redirect&);
	virtual ~Redirect ();

	static Redirect *clone (const Redirect&);

	bool active () const { return _active; }
	void set_active (bool yn, void *src) { _active = yn; active_changed (this, src); }

	unsigned long sort_key() const { return _sort_key; }
	void set_sort_key (unsigned long sk,void *src) { _sort_key = sk; sort_key_changed (this, src); }

	Placement placement() const { return _placement; }
	void set_placement (Placement, void *src);

	static string placement_as_string (Placement);

	virtual void run (vector<Sample *>& ibufs, jack_nframes_t nframes, jack_nframes_t offset) = 0;
	virtual void activate () = 0;
	virtual void deactivate () = 0;

	SigC::Signal2<void,Redirect*,void*> active_changed;
	SigC::Signal2<void,Redirect*,void*> sort_key_changed;
	SigC::Signal2<void,Redirect*,void*> placement_changed;
	SigC::Signal2<void,Redirect*,bool>  AutomationPlaybackChanged;
	SigC::Signal2<void,Redirect*,unsigned long> AutomationChanged;
	SigC::Signal1<void,Redirect*> GoingAway;

	static SigC::Signal1<void,Redirect*> RedirectCreated;

	XMLNode& get_state (void);
	int set_state (const XMLNode&);

	virtual UndoAction get_memento () const;
	void set_state (state_id_t);

	void *get_gui () const { return _gui; }
	void  set_gui (void *p) { _gui = p; }

	virtual string describe_parameter (unsigned long which);

	void add_automation_event (unsigned long port, float value, jack_nframes_t when);
	void reset_automation_event (unsigned long index, float value, jack_nframes_t when);
	void reset_automation_event_range (unsigned long start, unsigned long end, unsigned long what, double delta,
					   double lower, double upper);
	void remove_automation_event (unsigned long index);
	void remove_automation_events (unsigned long, unsigned long);
	void set_automation_playback (bool yn);
	bool get_automation_playback () const;

	int load_automation (string path);
	int save_automation (string path);
	void what_has_automation (set<unsigned long>&) const;

	AutomationList& automation_list() { return _automation_list; }
	
	enum FreezeType {
		None           = 0x0,
		AutomationData = 0x1,
	};

	void freeze (FreezeType);
	void thaw (FreezeType);

  protected:
	void set_placement (const string&, void *src);

	/* children may use this stuff as they see fit */

	AutomationList       _automation_list;
	mutable PBD::NonBlockingLock _automation_lock;
	bool                 _automation_playback;

  private:
	bool _active;
	Placement _placement;
	unsigned long _sort_key;
	void* _gui;  /* generic, we don't know or care what this is */

	void notify_about_automation_state ();

	FreezeType frozen;
	FreezeType pending;
	set<unsigned long>* pending_automation_data_notices;
};

}; /* namespace ARDOUR */

#endif /* __ardour_redirect_h__ */
