/*
    Copyright (C) 2000-2001 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: region.h,v 1.22 2003/02/18 17:59:23 pbd Exp $
*/

#ifndef __ardour_region_h__
#define __ardour_region_h__

#include <vector>

#include <pbd/fastlog.h>
#include <pbd/undo.h>

#include <ardour/ardour.h>
#include <ardour/source.h>
#include <ardour/gain.h>
#include <ardour/logcurve.h>

class XMLNode;

namespace ARDOUR {

class Route;
class Playlist;

struct RegionState {
    RegionState() {}
    
    jack_nframes_t      _start;
    jack_nframes_t      _length;
    jack_nframes_t      _position;
    unsigned short 	_flags;
    LogCurve       	_fade_in;
    LogCurve       	_fade_out;
    jack_nframes_t      _sync_offset;
    layer_t        	_layer;
    mutable Gain   	_envelope;
    string              _name;        
};

class Region : private RegionState, public Stateful, public SigC::Object {
  public:
	struct State {
	    string operation;
	    RegionState state;
	    State (const string& why) : operation (why) {}
	};

  private:
	typedef vector<State *> StateMap;

  public:

	enum Flag {
		Muted = 0x1,
		Opaque = 0x2,
		EnvelopeActive = 0x4,
		DefaultFadeIn = 0x8,
		DefaultFadeOut = 0x10,
		Locked = 0x20,
		Automatic = 0x40,
		WholeFile = 0x80,
		FadeIn = 0x100,
		FadeOut = 0x200,
		Copied = 0x400
		
	};

	// don't extend this without also working on FreezeIndex below

	enum FreezeFlag {
		BoundsNotification = 0x1,
		FadeNotification = 0x2,
		SyncNotification = 0x8,
		EnvelopeNotification = 0x10,
		LayerNotification = 0x20
	};

	enum BoundsChangeInfo {
		StartChange = 0x1,
		PositionChange = 0x2,
		LengthChange = 0x4
	};

	typedef vector<Source *> SourceList;
	
	static const Flag DefaultFlags = Flag (Opaque|DefaultFadeIn|DefaultFadeOut|FadeIn|FadeOut);

	/* This is emitted only when a new id is assigned. Therefore,
	   in a pure Region copy, it will not be emitted.
	*/

	static SigC::Signal1<void,Region*> RegionCreated;
	
	Region (Source&, jack_nframes_t start, jack_nframes_t length, const string& name, layer_t = 0, Flag flags = DefaultFlags);
	Region (SourceList &, jack_nframes_t start, jack_nframes_t length, const string& name, layer_t = 0, Flag flags = DefaultFlags);
	Region (const Region&, jack_nframes_t start, jack_nframes_t length, const string& name, layer_t = 0, Flag flags = DefaultFlags);
	Region (const Region&, const string& name);
	Region (const Region&);
	Region (Source&, const XMLNode&);
	Region (SourceList &, const XMLNode&);
	~Region();

	Source& source(unsigned int n=0) const { if (n < sources.size()) return *sources[n]; else return *sources[0]; } 
	Gain&   envelope() { return _envelope; }

	unsigned int n_channels() { return sources.size(); }
	
	ARDOUR::id_t id() const { return _id; }

	/* Note: changing the name of a Region does not constitute an edit */

	const string& name() const { return _name; }
	void set_name (const string& str);

	jack_nframes_t position () const { return _position; }
	jack_nframes_t start () const { return _start; }
	jack_nframes_t length() const { return _length; }
	jack_nframes_t sync_offset() const { return _sync_offset; }
	layer_t layer () const { return _layer; }
	
	/* first_frame() is an alias; last_frame() just hides some math */

	jack_nframes_t first_frame() const { return _position; }
	jack_nframes_t last_frame() const { return _position + _length - 1; }

	bool muted() const { return _flags & Muted; }
	bool opaque () const { return _flags & Opaque; }
	bool envelope_active () const { return _flags & EnvelopeActive; }
	bool locked() const { return _flags & Locked; }
	bool automatic() const { return  _flags & Automatic; }
	bool whole_file() const { return _flags & WholeFile ; }
	bool fade_in_active () const { return _flags & FadeIn; }
	bool fade_out_active () const { return _flags & FadeOut; }

	const LogCurve& fade_in() const;
	const LogCurve& fade_out() const;

	void freeze (FreezeFlag);
	void thaw (FreezeFlag, const string& why, bool with_sig, void *src);

	bool covers (jack_nframes_t frame) const {
		return _position <= frame && frame < _position + _length;
	}

	enum OverlapType {
		OverlapNone,
		OverlapInternal,
		OverlapStart,
		OverlapEnd,
		OverlapExternal,
	};

	OverlapType coverage (jack_nframes_t start, jack_nframes_t end) {
		jack_nframes_t rend = _position + _length - 1;

		if ((start >= _position) && (end <= rend)) {
			return OverlapInternal;
		}
		if ((end >= _position) && (end <= rend)) {
			return OverlapStart;
		}
		if ((start >= _position) && (start <= rend)) {
			return OverlapEnd;
		}
		if ((_position >= start) && (_position <= end) && (rend <= end)) {
			return OverlapExternal;
		}
		return OverlapNone;
	}

	virtual bool is_crossfade() const { return false; }

	// jack_nframes_t read (Sample *buf, jack_nframes_t offset, jack_nframes_t cnt) const;

	jack_nframes_t read_peaks (PeakData *buf, jack_nframes_t npeaks, jack_nframes_t offset, jack_nframes_t cnt, unsigned int chan_n=0) const;
	jack_nframes_t read_at (Sample *buf, Sample *mixdown_buffer, 
				float *gain_buffer, jack_nframes_t position, jack_nframes_t cnt, unsigned int chan_n=0) const;

	SigC::Signal2<void,BoundsChangeInfo,void*> BoundsChanged;
	SigC::Signal0<void> NameChanged;
	SigC::Signal0<void> MuteChanged;
	SigC::Signal0<void> OpacityChanged;
	SigC::Signal0<void> SyncOffsetChanged;
	SigC::Signal0<void> FadeInChanged;
	SigC::Signal0<void> FadeOutChanged;
	SigC::Signal0<void> FadeInActiveChanged;
	SigC::Signal0<void> FadeOutActiveChanged;
	SigC::Signal0<void> LockChanged;
	SigC::Signal0<void> EnvelopeChanged;
	SigC::Signal0<void> EnvelopeActiveChanged;
	SigC::Signal0<void> LayerChanged;
	SigC::Signal1<void,void*> SelectedChanged;

	SigC::Signal1<void,Region*> GoingAway;

	XMLNode& get_state ();
	int      set_state (const XMLNode&);
	XMLNode& get_short_state (); /* used only by Session */

	static void set_default_fade (float steepness, jack_nframes_t len);

	void use_default_fade_in ();
	void use_default_fade_out ();

	/* EDITING OPERATIONS */

	void set_length (jack_nframes_t, void *src);
	void set_start (jack_nframes_t, void *src);
	void set_position (jack_nframes_t, void *src);
	void set_position_and_layer (jack_nframes_t, layer_t, void *src);

	void nudge_position (long, void *src);

	void trim_start (jack_nframes_t new_position, void *src);
	void trim_front (jack_nframes_t new_position, void *src);
	void trim_end (jack_nframes_t new_position, void *src);
	void trim_to (jack_nframes_t position, jack_nframes_t length, void *src);
	
	void set_layer (layer_t l, int dir = 1);
	void set_layer_from_playlist (layer_t l);  /* ONLY Playlist can call this */
	void raise ();
	void lower ();
	void raise_to_top ();
	void lower_to_bottom ();

	void set_fade_in_active (bool yn);
	void set_fade_in (float steepness, jack_nframes_t len);
	void set_fade_in_length (jack_nframes_t);
	void set_fade_in_steepness (float);

	void set_fade_out_active (bool yn);
	void set_fade_out (float steepness, jack_nframes_t len);
	void set_fade_out_length (jack_nframes_t);
	void set_fade_out_steepness (float);

	void set_sync_offset (jack_nframes_t n);
	void set_muted (bool yn);
	void set_opaque (bool yn);
	void set_envelope_active (bool yn);
	void set_locked (bool yn);

	void add_envelope_control_point (jack_nframes_t, gain_t);
	void remove_envelope_control_point (size_t);
	void remove_envelope_control_points (size_t, size_t);
	void modify_envelope_control_point (size_t, jack_nframes_t, gain_t);
	void modify_envelope_control_points (size_t, size_t, gain_t);

	void set_state (state_id_t);

	ARDOUR::Playlist* get_playlist() const { return _playlist; }

	UndoAction get_memento() const;

  private:
	friend class Playlist;

	void save_state (const string &);
	
  private:
	SourceList        sources;
	
//	Source&              _source;
	PBD::Lock             lock;
	gain_t              *_fade_in_data;
	gain_t              *_fade_out_data;
	ARDOUR::id_t         _id;
	state_id_t           _current_state_id;
	bool                 _first_edit;
	ARDOUR::Playlist*    _playlist;

	enum FreezeIndex {
		BoundsNotificationIndex,
		FadeNotificationIndex,
		SyncNotificationIndex,
		EnvelopeNotificationIndex,
		LayerNotificationIndex,
		MaxFrozen // must be last
	};

	unsigned char        _frozen[MaxFrozen];

	bool hold_bounds_notification() const { return _frozen[BoundsNotificationIndex]; }
	bool hold_fade_notification() const { return _frozen[FadeNotificationIndex]; }
	bool hold_sync_notification() const { return _frozen[SyncNotificationIndex]; }
	bool hold_envelope_notification() const { return _frozen[EnvelopeNotificationIndex]; }
	bool hold_layer_notification() const { return _frozen[LayerNotificationIndex]; }
	
	void set_default_envelope ();

	static gain_t    *default_fade_in_data;
	static gain_t    *default_fade_out_data;
	static LogCurve   default_fade_in;
	static LogCurve   default_fade_out;

	static void build_fade (gain_t **gain, LogCurve&, bool in);

	StateMap states;
	void store_state (RegionState&) const;
	void restore_state (RegionState&);
	
	void trim_to_internal (jack_nframes_t position, jack_nframes_t length, void *src);
	void recompute_gain_at_end ();
	void recompute_gain_at_start ();

	void remove_space ();

	void set_playlist (ARDOUR::Playlist*);

	bool copied() const { return _flags & Copied; }
	void maybe_uncopy ();
};

} /* namespace ARDOUR */

/* access from C objects */

extern "C" {
	int    region_read_peaks_from_c   (void *arg, gulong npeaks, gulong start, gulong length, gpointer data, unsigned int n_chan);
	gulong region_length_from_c (void *arg);
}

#endif /* __ardour_region_h__ */
