/*
    Copyright (C) 2000-2002 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: route.h,v 1.42 2003/02/14 18:22:12 pbd Exp $
*/

#ifndef __ardour_route_h__
#define __ardour_route_h__

#include <iostream>
#include <list>
#include <set>
#include <map>

#include <pthread.h>
#include <glib.h>

#include <math.h>

#include <pbd/atomic.h>
#include <pbd/fastlog.h>
#include <pbd/lockmonitor.h>
#include <pbd/xml++.h>
#include <pbd/undo.h>

#include <ardour/ardour.h>
#include <ardour/stateful.h>
#include <ardour/io.h>
#include <ardour/session.h>
#include <ardour/curve.h>
#include <ardour/redirect.h>

namespace ARDOUR {

class Insert;
class Send;
class RouteGroup;

enum mute_type {
    PRE_FADER =    0x1,
    POST_FADER =   0x2,
    CONTROL_OUTS = 0x4,
    MAIN_OUTS =    0x8
};

class Route : public IO
{
  private:
	typedef list<Redirect *> RedirectList;

  public:
	Route (Session&, const string& name, int input_min, int input_max, int output_min, int output_max);
	Route (Session&, const XMLNode&);
	virtual ~Route();

	const string& comment() { return _comment; }
	void set_comment (const string &str, void *src);

	unsigned long order_key() const { return _order_key; }
	void set_order_key (unsigned long ok) {
		_order_key = ok;
	}

	bool hidden() const { return _flags & Hidden; }

	/* these are the core of the API of a Route. see the protected sections as well */

	virtual int  roll (jack_nframes_t nframes, jack_nframes_t end_frame, jack_nframes_t offset, int declick,
			   bool can_record, bool rec_monitors_input);
	virtual int  no_roll (jack_nframes_t nframes, jack_nframes_t end_frame, jack_nframes_t offset, bool state_changing,
			      bool can_record, bool rec_monitors_input);
	virtual void toggle_monitor_input ();
	virtual bool can_record() const { return false; }
	virtual void set_record_enable (bool yn, void *src) {}
	virtual bool record_enabled() const { return false; }
	virtual void transport_stopped (bool abort);
	virtual IO& input_io() { return *this; }
	virtual IO& output_io(){ return *this; }

	/* end of vfunc-based API */

	bool active() const { return _active; }
	void set_active (bool yn);

	void set_solo (bool yn, void *src);
	bool soloed() const { return _soloed; }

	void set_solo_safe (bool yn, void *src);
	bool solo_safe() const { return _solo_safe; }

	void set_mute (bool yn, void *src);
	bool muted() const { return _muted; }

	void set_mute_config (mute_type, bool, void *src);
	bool get_mute_config (mute_type);

	void set_phase_invert (bool yn, void *src);
	bool phase_invert() const { return _phase_invert; }
	
	void       set_edit_group (RouteGroup *, void *);
	RouteGroup *edit_group () { return _edit_group; }

	void       set_mix_group (RouteGroup *, void *);
	RouteGroup *mix_group () { return _mix_group; }

	/* these overrides IO::set_gain for automation reasons */

	void set_gain (gain_t val, void *src);
	void inc_gain (gain_t delta, void *src);

	gain_t effective_gain() const;
	pan_t effective_stereo_pan() const;

	void  set_meter_pre_fader (bool yn) { _meter_pre_fader = yn;}
	bool  meter_pre_fader() const { return _meter_pre_fader; }

	/* Redirects */

	template<class T> void foreach_redirect (T *obj, void (T::*func)(Redirect *)) {
		LockMonitor lm (redirect_lock, __LINE__, __FILE__);
		for (RedirectList::iterator i = _redirects.begin(); i != _redirects.end(); i++) {
			(obj->*func) (*i);
		}
	}

	Redirect *nth_redirect (unsigned long n) {
		LockMonitor lm (redirect_lock, __LINE__, __FILE__);
		RedirectList::iterator i;
		for (i = _redirects.begin(); i != _redirects.end() && n; i++, n--);
		if (i == _redirects.end()) {
			return 0;
		} else {
			return *i;
		}
	}
		

	void add_redirect (Redirect *, void *src);
	void move_redirect (Redirect *, void *src);
	void remove_redirect (Redirect *, void *src);
	void clear_redirects (void *src);
	void sort_redirects ();
	void copy_redirects (const Route&, Redirect::Placement);
	void all_redirects_flip();
	void all_redirects_active (bool state);

	SigC::Signal1<void,void*> solo_changed;
	SigC::Signal1<void,void*> solo_safe_changed;
	SigC::Signal1<void,void*> comment_changed;
	SigC::Signal1<void,void*> mute_changed;
	SigC::Signal1<void,void*> pre_fader_changed;
	SigC::Signal1<void,void*> post_fader_changed;
	SigC::Signal1<void,void*> control_outs_changed;
	SigC::Signal1<void,void*> main_outs_changed;
	SigC::Signal1<void,void*> redirects_changed;
	SigC::Signal1<void,void*> record_enable_changed;
	SigC::Signal1<void,void*> edit_group_changed;
	SigC::Signal1<void,void*> mix_group_changed;
	SigC::Signal2<void,Redirect*,unsigned long> redirect_automation_changed;
	SigC::Signal0<void>       active_changed;

	SigC::Signal0<void> GoingAway;

	/* gui's call this for their own purposes. */

	SigC::Signal2<void,string,void*> gui_changed;

	/* stateful */

	XMLNode& get_state();
	int set_state(const XMLNode& node);
	XMLNode& get_template();

	SigC::Signal1<void,void*> SelectedChanged;

	/* undo */

	UndoAction get_memento() const;
	void set_state (state_id_t);

	/* automation */

	Curve& gain_automation_curve () { return _gain_automation_curve; }
	Curve& pan_automation_curve () { return _pan_automation_curve; }
	
	void add_automation_event (AutomationType, jack_nframes_t when, float value);
	void add_automation_event (const string&, jack_nframes_t when, float value);
	void remove_automation_events (AutomationType, jack_nframes_t first_frame, jack_nframes_t nframes);
	void remove_automation_events (const string&, jack_nframes_t first_frame, jack_nframes_t nframes);

	static void set_automation_interval (jack_nframes_t frames) {
		_automation_interval = frames;
	}

	static jack_nframes_t automation_interval() { 
		return _automation_interval;
	}

	void clear_automation ();

	void set_automation_recording (bool yn, void *src);
	void set_automation_playback (bool yn, void *src);

	bool automation_recording() { return _automation_recording; }
	bool automation_playback() { return  _automation_playback; }

	void set_automation_recording_bits (unsigned long flags);
	void set_automation_playback_bits (unsigned long flags);
	unsigned long automation_recording_bits() { return _automation_recording_bits; }
	unsigned long automation_playback_bits() { return _automation_playback_bits; }

	SigC::Signal1<void,void*> automation_recording_changed;
	SigC::Signal1<void,void*> automation_playback_changed;

	int set_control_outs (const vector<string>& ports);
	IO* control_outs() { return _control_outs; }

	bool feeds (Route *);
	set<Route *> fed_by;

  protected:
	friend class Session;
	Session& session;

	void set_solo_mute (bool yn);
	void set_block_size (jack_nframes_t nframes);
	bool has_external_redirects() const;

  protected:
	/* for derived classes */

	enum Flag {
		Hidden = 0x1
	};

	unsigned char _flags;

  protected:

	/* tight cache-line access here is more important than sheer speed of
	   access.
	*/

	bool                     _muted : 1;
	bool                     _soloed : 1;
	bool                     _solo_muted : 1;
	bool                     _solo_safe : 1;
	bool                     _phase_invert : 1;
	bool                     _recordable : 1;
	bool                     _active : 1;
	bool                     _mute_affects_pre_fader : 1;
	bool                     _mute_affects_post_fader : 1;
	bool                     _mute_affects_control_outs : 1;
	bool                     _mute_affects_main_outs : 1;
	bool                     _silent : 1;
	bool                     _meter_pre_fader;

	gain_t                    solo_gain;
	gain_t                    mute_gain;
	gain_t                    desired_solo_gain;
	gain_t                    desired_mute_gain;

	RedirectList             _redirects;
	PBD::NonBlockingLock      redirect_lock;
	IO                      *_control_outs;
	PBD::NonBlockingLock      control_outs_lock;
	RouteGroup              *_edit_group;
	RouteGroup              *_mix_group;
	string                   _comment;
	jack_nframes_t            block_size;
	
	void get_process_buffers (vector<Sample*>& bufs, jack_nframes_t nframes, jack_nframes_t offset);
	virtual void passthru     (jack_nframes_t nframes, jack_nframes_t offset, int declick);
	void process_output_buffers (vector<Sample*>& bufs, jack_nframes_t nframes, jack_nframes_t offset, bool with_redirects, int declick);
	
  protected:
	/* for derived classes */

	virtual XMLNode& state(bool);

	void silence (jack_nframes_t nframes, jack_nframes_t offset);
	SigC::Connection input_signal_connection;

	jack_nframes_t last_automation_snapshot;
	static jack_nframes_t _automation_interval;

	void automation_snapshot (jack_nframes_t now) {
		if (_automation_recording) {
			if (last_automation_snapshot > now || now - last_automation_snapshot > _automation_interval) {
				if (_automation_recording_bits & GainAutomation) {
					add_automation_event (GainAutomation, now, gain());
				}
				if (_automation_recording_bits & PanAutomation) {
					/* we only automate stereo pan right now */
					if (n_outputs() == 2) {
						add_automation_event (PanAutomation, now, pans()[0]);
					}
				}
				last_automation_snapshot = now;
			}
		}
	}

	bool    _automation_recording;
	unsigned long _automation_recording_bits;
	bool    _automation_playback;
	unsigned long _automation_playback_bits;

	bool     apply_gain_automation;
	bool     apply_pan_automation;

	Curve    _gain_automation_curve;
	Curve    _pan_automation_curve;
	
	gain_t *gain_automation_buffer;
	pan_t  *pan_automation_buffer;

	int  save_automation (const string&);
	int  load_automation (const string&);
	
	PBD::NonBlockingLock automation_lock;

	state_id_t _current_state_id;

  private:
	gain_t _effective_gain;
	gain_t _effective_stereo_pan;

	void init ();

	static unsigned long order_key_cnt;
	unsigned long _order_key;

	void input_change_handler (void *src);
	void output_change_handler (void *src);
};

}; /* namespace ARDOUR*/

#endif /* __ardour_route_h__ */
