/*
    Copyright (C) 2000 Paul Davis 

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    $Id: session.h,v 1.69 2003/02/21 15:29:50 pbd Exp $
*/

#ifndef __ardour_session_h__
#define __ardour_session_h__

#include <string>
#if __GNUC__ >= 3
#include <ext/slist>
using __gnu_cxx::slist;
#else
#include <slist.h>
#endif
#include <map>
#include <vector>
#include <set>
#include <stack>

#include <glib.h>
#include <sndfile.h>

#include <pbd/error.h>
#include <pbd/lock_free_fifo.h>
#include <pbd/atomic.h>
#include <pbd/lockmonitor.h>
#include <pbd/undo.h>
#include <pbd/pool.h>

#include <midi++/types.h>
#include <midi++/mmc.h>

#include <ardour/ardour.h>
#include <ardour/configuration.h>
#include <ardour/location.h>
#include <ardour/stateful.h>
#include <ardour/region.h>
#include <ardour/playlist.h>
#include <ardour/gain.h>
#include <ardour/io.h>

class XMLTree;
class XMLNode;

namespace MIDI {
	class Port;
}

namespace ARDOUR {

class Port;
class AudioEngine;
class Slave;
class DiskStream;	
class Route;
class AuxInput;
class Source;
class FileSource;
class Auditioner;
class Insert;
class Send;
class Redirect;
class PortInsert;
class PluginInsert;
class Connection;
class TempoMap;
class AudioTrack;
class NamedSelection;

struct AudioExportSpecification;
struct RouteGroup;

using std::vector;
using std::string;

class Session : public SigC::Object, public Stateful

{
  private:
	typedef pair<Route*,bool> RouteBooleanState;
	typedef vector<RouteBooleanState> GlobalRouteBooleanState;

  public:
	enum RecordState {
		Disabled = 0,
		Enabled = 1,
		Recording = 2
	};

	enum State { 
		Stopped = 0,
		Roll,
		Forward,
		Rewind,
		Locating
	};
	
	struct Event {
	    enum Type {
		    TransportState,
		    PunchIn,
		    PunchOut,
		    AutoStop,
		    AutoLoop
	    };

	    enum Action {
		    Add,
		    Remove,
		    Replace
	    };

	    Type		type;
	    Action              action;
	    jack_nframes_t	action_frame;
	    jack_nframes_t	target_frame;
	    Session::State      transport_state;
	    bool             abort; /* for transport stop/capture only */

	    Event () {
		    action = Add;
		    action_frame = 0;
		    target_frame = 0;
		    abort = false;
	    }

	    bool before (const Event& other) const {
		    return action_frame < other.action_frame;
	    }

	    bool after (const Event& other) const {
		    return action_frame > other.action_frame;
	    }

	    static bool compare (const Event *e1, const Event *e2) {
		    return e1->before (*e2);
	    }
	};

	Session (AudioEngine&,
		 MIDI::Port *mmcp,
		 MIDI::Port *mtcp,
		 string fullpath,
		 string snapshot_name,
		 string* mix_template = 0);
	
	virtual ~Session ();

	static string find_session (string name, bool& isnew);
	
	string path() const { return _path; }
	string name() const { return _name; }

	void set_dirty () {
		_state_of_the_state = StateOfTheState (_state_of_the_state | Dirty);
	}

	bool dirty() const { return _state_of_the_state & Dirty; }

	string sound_dir () const;
	string automation_dir () const;

	static string template_dir ();
	
	static string peak_path_from_audio_path (string);

	void process (jack_nframes_t nframes);

	const vector<Sample*>& get_passthru_buffers() const { return _passthru_buffers; }

	DiskStream    *diskstream_by_id (unsigned long id);
	DiskStream    *diskstream_by_name (string name);

	void refill_all_diskstream_buffers ();
	void refill_diskstream_buffer (DiskStream&);
	void set_diskstream_speed (DiskStream&, float speed, void *src);
	unsigned long diskstream_buffer_size() const { return dstream_buffer_size; }
	unsigned long get_next_diskstream_id() const { return n_diskstreams(); }
	unsigned long n_diskstreams() const;
	
	typedef list<DiskStream *> DiskStreamList;

	Session::DiskStreamList disk_streams() const {
		LockMonitor lm (diskstream_lock, __LINE__, __FILE__);
		return diskstreams; /* XXX yes, force a copy */
	}

	void foreach_diskstream (void (DiskStream::*func)(void));
	template<class T> void foreach_diskstream (T *obj, void (T::*func)(DiskStream&));

	typedef slist<Route *> RouteList;

	RouteList get_routes() const {
		LockMonitor rlock (route_lock, __LINE__, __FILE__);
		return routes; /* XXX yes, force a copy */
	}

	struct RoutePublicOrderSorter {
	    bool operator() (Route *, Route *b);
	};
	
	template<class T> void foreach_route (T *obj, void (T::*func)(Route&));

	Route *route_by_name (string);

	State transport_state () const { return _transport_state; }

	bool get_record_enabled() const { 
		return (record_status () >= Enabled);
	}

	RecordState record_status() const {
		return (RecordState) atomic_read (&_record_status);
	}

	bool actively_recording () {
		return record_status() == Recording;
	}
	
	SigC::Signal0<void> going_away;

	/* Record status signals */

	SigC::Signal0<void> RecordEnabled;
	SigC::Signal0<void> RecordDisabled;

	/* Transport mechanism signals */

	SigC::Signal0<void> TransportStateChange; /* generic */
	SigC::Signal1<void,jack_nframes_t> PositionChanged; /* sent after any non-sequential motion */
	SigC::Signal0<void> DurationChanged;

	SigC::Signal1<void,Route&> RouteAdded;
	SigC::Signal1<void,DiskStream&> DiskStreamAdded;

	void request_transport_state (Session::State, bool abort = false);
	void request_locate (jack_nframes_t frame);
	void goto_end ()   { request_locate (current_end);}
	void goto_start () { request_locate (0); }
	void use_rf_shuttle_speed ();

	int wipe ();
	int wipe_diskstream (DiskStream *);

	jack_nframes_t current_end_frame() const { return current_end; }
	unsigned long nstreams () const      { return dstream_cnt; }
	unsigned long nroutes () const       { return route_cnt; }
	jack_nframes_t frame_rate() const   { return _current_frame_rate; }
	jack_nframes_t frames_per_smpte_frame() const { return _frames_per_smpte_frame; }

	/* Range play */

	void set_auto_play_range (jack_nframes_t start, jack_nframes_t end);

	/* Locations */

	Locations *locations() { return &_locations; }

	SigC::Signal1<void,Location*>    auto_loop_location_changed;
	SigC::Signal1<void,Location*>    auto_punch_location_changed;
	SigC::Signal0<void>              locations_modified;

	void set_auto_punch_location (Location *);
	void set_auto_loop_location (Location *);

	bool record_enabling_legal () const;
	
	void maybe_enable_record ();
	void disable_record () {
		atomic_set (&_record_status, Disabled);
		send_mmc_in_another_thread (MIDI::MachineControl::cmdRecordExit);
		RecordDisabled ();
	}

	enum ControlType {
		AutoPlay,
		AutoLoop,
		AllSafe,
		DynamicPunch,
		AutoInput,
		AutoPunch,
		SendMTC,
		MMCControl,
		Live,
		RecordingPlugins,
		AutoCrossFade,
		SendMMC,
		SlaveType,
		Clicking
	};

	SigC::Signal1<void,ControlType> ControlChanged;

	void set_auto_play (bool yn) { auto_play = yn; ControlChanged (AutoPlay); }
	void set_auto_loop (bool yn);
	void set_all_safe (bool yn) { all_safe = yn; ControlChanged (AllSafe); }
	void set_dynamic_punch (bool yn) { dynamic_punch = yn; ControlChanged (DynamicPunch); }
	void set_auto_input (bool yn);
	void set_auto_punch (bool yn);
	void set_send_mtc (bool yn);
	void set_send_mmc (bool yn);
	void set_mmc_control (bool yn);
	void set_recording_plugins (bool yn) { recording_plugins = yn; ControlChanged (RecordingPlugins); }
	void set_auto_crossfade (bool yn) { auto_crossfade = yn; ControlChanged (AutoCrossFade); }

	bool get_auto_play () const { return auto_play; }
	bool get_auto_input () const { return auto_input; }
	bool get_auto_loop () const { return auto_loop; }
	bool get_auto_punch () const { return auto_punch; }
	bool get_all_safe () const { return all_safe; }
	bool get_dynamic_punch () const { return dynamic_punch; }
	bool get_send_mtc () const;
	bool get_send_mmc () const;
	bool get_mmc_control () const { return mmc_control; }
	bool get_recording_plugins () const { return recording_plugins; }
	bool get_auto_crossfade () const { return auto_crossfade; }

	void add_event (jack_nframes_t action_frame, Event::Type type, jack_nframes_t target_frame = 0);
	void remove_event (jack_nframes_t frame, Event::Type type);

	bool xrun_recoverable ();
	jack_nframes_t get_block_size() const { return current_block_size; }

	int save_state (string snapshot_name);
	int restore_state (string snapshot_name);
	int save_as (string new_name);
	int save_template (string template_name);
	static int rename_template (string old_name, string new_name);
	static int delete_template (string name);
	
	SigC::Signal1<void,string> StateSaved;

	vector<string*>* possible_states() const;
	static vector<string*>* possible_states(string path);

	XMLNode& get_state();
	int      set_state(const XMLNode& node);
	XMLNode& get_template();

	void swap_configuration(Configuration** new_config);
	void copy_configuration(Configuration* new_config);

	enum StateOfTheState {
		Clean = 0x0,
		Dirty = 0x1,
		CannotSave = 0x2,
		Deletion = 0x4,
		InitialConnecting = 0x8
	};

	StateOfTheState state_of_the_state() const { return _state_of_the_state; }

	RouteGroup* add_edit_group (string);
	RouteGroup* add_mix_group (string);

	RouteGroup *mix_group_by_name (string);
	RouteGroup *edit_group_by_name (string);

	SigC::Signal1<void,RouteGroup*> edit_group_added;
	SigC::Signal1<void,RouteGroup*> mix_group_added;

	template<class T> void foreach_edit_group (T *obj, void (T::*func)(RouteGroup *)) {
		list<RouteGroup *>::iterator i;
		for (i = edit_groups.begin(); i != edit_groups.end(); i++) {
			(obj->*func)(*i);
		}
	}

	template<class T> void foreach_mix_group (T *obj, void (T::*func)(RouteGroup *)) {
		list<RouteGroup *>::iterator i;
		for (i = mix_groups.begin(); i != mix_groups.end(); i++) {
			(obj->*func)(*i);
		}
	}

	/* fundamental operations. duh. */

	AudioTrack *new_audio_track ();
	Route *new_audio_route ();

	void   remove_route (Route&);
	void   resort_routes (void *src);

	AudioEngine &engine() { return _engine; };

	/* configuration. there should really be accessors/mutators
	   for these 
	*/

	float   meter_hold;
	gint32  max_level;
	gint32  min_level;
	string  click_emphasis_sound;
	string  click_sound;
	bool    click_requested;
	jack_nframes_t over_length_short;
	jack_nframes_t over_length_long;
	bool    send_midi_timecode;
	bool    send_midi_machine_control;
	float   shuttle_speed_factor;
	float   shuttle_speed_threshold;
	float   rf_speed;
	float   smpte_frames_per_second;
	bool    smpte_drop_frames;

	/* Time */

	jack_nframes_t transport_frame () const {return _transport_frame; }
	jack_nframes_t audible_frame () const;

	int  set_smpte_type (float fps, bool drop_frames);

	void bbt_time (jack_nframes_t when, BBT_Time&);

	void smpte_time (SMPTE_Time &);
	void smpte_time (jack_nframes_t when, SMPTE_Time&);
	void smpte_time_subframes (jack_nframes_t when, SMPTE_Time&);

	void smpte_duration (jack_nframes_t, SMPTE_Time&) const;
	void smpte_duration_string (char *, jack_nframes_t) const;

	void           set_smpte_offset (jack_nframes_t);
	jack_nframes_t smpte_offset (jack_nframes_t) const { return _smpte_offset; }

	SigC::Signal0<void> SMPTEOffsetChanged;

	enum SlaveSource {
		None = 0,
		MTC,
		ADAT,
		JACK,
		Scrub
	};

	int         set_slave_source (SlaveSource);
	SlaveSource slave_source() const;
   	float       slave_speed() const { return _slave_speed; }
	float       current_speed() const { return _slave_speed; }

	TempoMap& tempo_map() { return *_tempo_map; }
	
	/* region info  */

	SigC::Signal1<void,Region *> RegionAdded;
	SigC::Signal1<void,Region *> RegionRemoved;

	int region_name (string& result, string base = string(""), bool newlevel = false) const;
	static string region_name_from_path (string path);
	string path_from_region_name (string name, string identifier);

	Region *XMLRegionFactory (const XMLNode&, bool full);

	template<class T> void foreach_region (T *obj, void (T::*func)(Region *)) {
		LockMonitor lm (region_lock, __LINE__, __FILE__);
		for (RegionList::iterator i = regions.begin(); i != regions.end(); i++) {
			(obj->*func) ((*i).second);
		}
	}

	/* source management */

	struct import_status {
	    volatile bool  done;
	    volatile bool  cancel;
	    volatile float percent_done;
	    string doing_what;
	};

	int import_audiofile (string path, import_status&);

	int export_audio (ARDOUR::AudioExportSpecification&);

	void add_source (Source *);

	SigC::Signal1<void,Source *> SourceAdded;
	SigC::Signal1<void,Source *> SourceRemoved;

	FileSource *create_file_source (ARDOUR::DiskStream&);
	Source *get_source (ARDOUR::id_t);

	/* playlist management */

	Playlist* playlist_by_name (string name);
	void add_playlist (Playlist *);
	SigC::Signal1<void,Playlist*> PlaylistAdded;
	SigC::Signal1<void,Playlist*> PlaylistRemoved;

	Playlist *get_playlist (string name);

	unsigned long n_playlists() const { 
		LockMonitor lm (playlist_lock, __LINE__, __FILE__);
		return playlists.size();
	}

	template<class T> void foreach_playlist (T *obj, void (T::*func)(Playlist *));

	/* named selections */

	NamedSelection* named_selection_by_name (string name);
	void add_named_selection (NamedSelection *);
	void remove_named_selection (NamedSelection *);

	template<class T> void foreach_named_selection (T& obj, void (T::*func)(NamedSelection&));
	SigC::Signal0<void> NamedSelectionAdded;
	SigC::Signal0<void> NamedSelectionRemoved;

	/* fade curves */

	float get_default_fade_length () const { return default_fade_msecs; }
	float get_default_fade_steepness () const { return default_fade_steepness; }
	void set_default_fade (float steepness, float msecs);

	/* auditioning */

	void audition_region (Region&);
	void audition_file (string path);
	void cancel_audition ();
	bool is_auditioning () const;
	void request_audition_cancel ();

	SigC::Signal1<void,bool> AuditionActive;

	/* flattening stuff */

	int flatten_whole_track (AudioTrack&);
	int flatten_track (AudioTrack&, jack_nframes_t start, jack_nframes_t cnt);
	int flatten ();

	/* session-wide solo/mute/rec-enable */

	void set_all_solo (bool);
	void set_all_mute (bool);
	
	void record_disenable_all ();
	void record_enable_all ();

	/* insert/send management */
	
	unsigned long n_port_inserts() const { return _port_inserts.size(); }
	unsigned long n_plugin_inserts() const { return _plugin_inserts.size(); }
	unsigned long n_sends() const { return _sends.size(); }

	string next_send_name();
	string next_insert_name();
	
	/* s/w "RAID" management */
	
	jack_nframes_t available_capture_duration();

	/* I/O Connections */

	template<class T> void foreach_connection (T *obj, void (T::*func)(Connection *));
	void add_connection (Connection *);
	void remove_connection (Connection *);
	Connection *connection_by_name (string) const;

	SigC::Signal1<void,Connection *> ConnectionAdded;
	SigC::Signal1<void,Connection *> ConnectionRemoved;

	/* MIDI */
	
	int set_mtc_port (string port_tag);
	int set_mmc_port (string port_tag);
	MIDI::Port *mtc_port() const { return _mtc_port; }
	MIDI::Port *mmc_port() const { return _mmc_port; }

	/* Scrubbing */

	void start_scrub (jack_nframes_t where);
	void stop_scrub ();
	void set_scrub_speed (float);

	/* History (for editors, mixers, UIs etc.) */

	void undo (unsigned int n) {
		history.undo (n);
	}
	void redo (unsigned int n) {
		history.redo (n);
	}

	unsigned int undo_depth() const { return history.undo_depth(); }
	unsigned int redo_depth() const { return history.redo_depth(); }
	string next_undo() const { return history.next_undo(); }
	string next_redo() const { return history.next_redo(); }

	void begin_reversible_command (string cmd_name, UndoAction *private_undo = 0);
	void commit_reversible_command (UndoAction* private_redo = 0);

	void add_undo (const UndoAction& ua) {
		current_cmd.add_undo (ua);
	}
	void add_redo (const UndoAction& ua) {
		current_cmd.add_redo (ua);
	}
	void add_redo_no_execute (const UndoAction& ua) {
		current_cmd.add_redo_no_execute (ua);
	}

	UndoAction global_solo_memento (void *src);
	UndoAction global_mute_memento (void *src);
	UndoAction global_record_enable_memento (void *src);

	/* edit mode */

	void set_edit_mode (Playlist::EditMode);

	/* clicking */

	void set_clicking (bool yn);
	bool get_clicking() const { return _clicking; }

	void set_click_sound (string path);
	void set_click_emphasis_sound (string path);
		
	/* tempo FX */

	struct TimeStretchRequest {
	    ARDOUR::Route&  route;
	    ARDOUR::Region& region;
	    jack_nframes_t  start;
	    jack_nframes_t  new_length;
	    float           progress;
	    bool            running;
	    bool            quick_seek;
	    bool            antialias;

	    TimeStretchRequest (ARDOUR::Route& ro, ARDOUR::Region& re) 
		    : route (ro), region (re) {}
	};

	Region* tempoize_region (TimeStretchRequest&);

	string raid_path() const { return _raid_path; }
	void   set_raid_path(string);

	/* need to call this whenever we change native file formats */

	void reset_native_file_format();

  protected:
	friend class AudioEngine;
	void set_block_size (jack_nframes_t nframes);
	void set_frame_rate (jack_nframes_t nframes);

  protected:
	friend class DiskStream;
	void stop_butler ();
	void wait_till_butler_finished();

  protected:
	friend class Route;

  private:
	enum SubState {
		PendingDeclickIn   = 0x1,
		PendingDeclickOut  = 0x2,
		StopPendingCapture = 0x4,
		AfterDeclickOut    = 0x8
	};

	/* stuff used in process() should be close together to
	   maximise cache hits
	*/

	typedef void (Session::*process_function_type)(jack_nframes_t);

	AudioEngine            &_engine;
	atomic_t                 processing_prohibited;
	process_function_type    process_function;
	process_function_type    last_process_function;
	jack_nframes_t          _current_frame_rate;
	State                   _transport_state;
	int                      transport_sub_state;
	atomic_t                _record_status;
	jack_nframes_t          _transport_frame;
	jack_nframes_t           current_end;
	Slave                  *_slave;
	float                   _slave_speed;
	volatile float          _desired_slave_speed;
	float                   _last_slave_speed;
	jack_nframes_t          _last_slave_transport_frame;
	jack_nframes_t           maximum_output_latency;
	jack_nframes_t           last_stop_frame;
	vector<Sample *>        _passthru_buffers;

	void ensure_passthru_buffers (unsigned long howmany);
	
	// stack<process_function_type> process_function_stack;
	PBD::NonBlockingLock  process_function_lock;

	void set_process_function (process_function_type);
	void restore_process_function ();

	void process_slave          (jack_nframes_t);
	void process_scrub          (jack_nframes_t);
	void process_without_events (jack_nframes_t);
	void process_with_events    (jack_nframes_t);
	void process_audition       (jack_nframes_t);

	void prepare_diskstreams ();
	void commit_diskstreams (jack_nframes_t, bool& session_requires_butler);
	int  process_routes (jack_nframes_t, jack_nframes_t);

	bool get_rec_monitors_input () {
		if (actively_recording()) {
			return true;
		} else {
			if (auto_input) {
				return false;
			} else {
				return true;
			}
		}
	}

	int get_transport_declick_required () {

		if (transport_sub_state & PendingDeclickIn) {
			transport_sub_state &= ~PendingDeclickIn;
			return 1;
		} else if (transport_sub_state & PendingDeclickOut) {
			return -1;
		} else {
			return 0;
		}
	}

	State get_current_transport_state () {
		if (state_changing()) {
			
			/* if we declicked (w/ fade out) and are now changing state,
			   pretend that we are stopped.
			*/
			
			if (transport_sub_state & AfterDeclickOut) {
				return Stopped;
			} else {
				return last_transport_state;
			}
		} else {
			return _transport_state;
		}
	}

	void maybe_stop (jack_nframes_t limit) {
		if (_transport_frame >= limit || _transport_frame == 0) {
			do_set_transport_state_in_another_thread (Stopped);
		}
	}

	void check_declick_out () {
		if (transport_sub_state & PendingDeclickOut) {
			/* We've declicked, so now move to whatever state we were heading for,
			   and note that we are supposed to be silent. We have to do this
			   because the transport state change could take a while (i.e. it
			   won't be complete before the next time process() is called
			*/
			transport_sub_state |= AfterDeclickOut;
			do_set_transport_state_in_another_thread (pending_transport_state, pending_transport_frame, pending_transport_abort);
		}
	}

	MIDI::MachineControl*    mmc;
	MIDI::Port*             _mmc_port;
	MIDI::Port*             _mtc_port;
	string                  _path;
	string                  _name;
	unsigned long            dstream_cnt;
	unsigned long            route_cnt;
	bool                     recording_plugins;
	State                    pending_transport_state;
	jack_nframes_t           pending_transport_frame;
	bool                     pending_transport_abort;
	State                    last_transport_state;

	/* toggles */

	bool auto_play;
	bool auto_punch;
	bool auto_loop;
	bool auto_looped;
	bool auto_input;
	bool auto_crossfade;
	bool all_safe;
	bool dynamic_punch;
	bool monitor_in;
	bool send_mtc;
	bool send_mmc;
	bool mmc_control;

	LockFreeFIFO<Event*> pending_events;

	void hookup_io ();
	void when_engine_running ();
	void graph_reordered ();

	string _current_snapshot_name;

	XMLTree* state_tree;
	StateOfTheState _state_of_the_state;

	void     auto_save();
	int      load_options (const XMLNode&);
	XMLNode& get_options () const;
	int load_state (string snapshot_name);

	void queue_event (Event *ev);

	jack_nframes_t   _last_roll_location;
	jack_nframes_t   _last_record_location;
	bool             roll_after_next_locate;

	Sample*          butler_mixdown_buffer;
	float*           butler_gain_buffer;
	pthread_t        butler_thread;
	PBD::Lock        butler_request_lock;
	pthread_cond_t   butler_request;
	pthread_cond_t   butler_paused;
	bool             butler_should_run;

	void             summon_butler ();
	int              start_butler_thread ();
	void             terminate_butler_thread ();
	static void    *_butler_thread_work (void *arg);
	void             butler_thread_work ();

	float            motion_speed;
	unsigned long          cumulative_rf_motion;
	unsigned long          rf_scale;

	void set_motion_speed (float speed);
	void set_rf_speed (float speed);
	void reset_rf_scale (jack_nframes_t frames_moved);

	Locations        _locations;
	void              locations_changed ();
	void              handle_locations_changed (Locations::LocationList&);

	Location        *auto_loop_location;
	Location        *auto_punch_location;

	SigC::Connection auto_punch_start_changed_connection;
	SigC::Connection auto_punch_end_changed_connection;
	void             auto_punch_start_changed (Location *);
	void             auto_punch_end_changed (Location *);

	SigC::Connection auto_loop_start_changed_connection;
	SigC::Connection auto_loop_end_changed_connection;
	void             auto_loop_start_changed (Location *);
	void             auto_loop_end_changed (Location *);

	typedef slist<Event *> Events;
	Events           events;
	Events::iterator next_event;

	/* there can only ever be one of each of these */

	Event *auto_loop_event;
	Event *punch_out_event;
	Event *punch_in_event;

	/* events */

	void dump_events () const;
	void merge_event (Event *);
	void replace_event (Event::Type, jack_nframes_t action_frame, jack_nframes_t target = 0);
	void _replace_event (Event::Type, jack_nframes_t action_frame, jack_nframes_t target);
	void _remove_event (jack_nframes_t action_frame, Event::Type);

	void first_stage_init (string name, string snapshot_name);
	int  second_stage_init (bool new_tracks);
	void find_current_end ();
	void find_current_end_unlocked ();
	int create (bool& new_session, string* mix_template);
	void remove_empty_sounds ();

	void setup_midi_control ();
	int  midi_read (MIDI::Port *);

	void enable_record () {
		atomic_set (&_record_status, Recording);
		_last_record_location = _transport_frame;
		send_mmc_in_another_thread (MIDI::MachineControl::cmdRecordStrobe);
		RecordEnabled ();
	}
	
	void increment_transport_position (unsigned int val) {
		if (max_frames - val < _transport_frame) {
			_transport_frame = max_frames;
		} else {
			_transport_frame += val;
		}
	}

	void decrement_transport_position (unsigned int val) {
		if (val < _transport_frame) {
			_transport_frame -= val;
		} else {
			_transport_frame = 0;
		}
	}

	void post_transport_motion ();
	static void *session_loader_thread (void *arg);

	void *do_work();

	void set_next_event ();
	int  process_event (Event *);

	/* MIDI Machine Control */

	void deliver_mmc (MIDI::MachineControl::Command, jack_nframes_t);

	void mmc_deferred_play (MIDI::MachineControl &);
	void mmc_stop (MIDI::MachineControl &);
	void mmc_record_strobe (MIDI::MachineControl &);
	void mmc_record_exit (MIDI::MachineControl &);
	void mmc_track_record_status (MIDI::MachineControl &, 
				      unsigned long track, bool enabled);
	void mmc_fast_forward (MIDI::MachineControl &);
	void mmc_rewind (MIDI::MachineControl &);
	void mmc_locate (MIDI::MachineControl &, const MIDI::byte *);
	void mmc_shuttle (MIDI::MachineControl &mmc, float speed, bool forw);
	void mmc_record_enable (MIDI::MachineControl &mmc, size_t track, bool enabled);

	MIDI::byte mmc_buffer[32];
	MIDI::byte mtc_msg[16];
	MIDI::byte mtc_smpte_bits;   /* encoding of SMTPE type for MTC */
	jack_nframes_t  last_mtc_smpte_frame;
	jack_nframes_t  last_outbound_mtc_frame;   

	jack_nframes_t _frames_per_smpte_frame;
	jack_nframes_t _smpte_offset;

	/* cache the most-recently requested time conversions.
	   this helps when we have multiple clocks showing the
	   same time (e.g. the transport frame)
	*/

	bool       last_smpte_valid;
	jack_nframes_t  last_smpte_when;
	SMPTE_Time last_smpte;

	int send_full_time_code ();
	int send_midi_time_code ();

	void send_full_time_code_in_another_thread ();
	void send_midi_time_code_in_another_thread ();
	void send_time_code_in_another_thread (bool full);
	void send_mmc_in_another_thread (MIDI::MachineControl::Command, jack_nframes_t frame = 0);

	jack_nframes_t adjust_apparent_position (jack_nframes_t frames);
	
	void reset_record_status ();
	
	/* These must be ONLY be called by the audio thread. No
	   run-time checking of this is done, however - I rely
	   on my own design skills ;)
	*/

	void audio_lock_transport ();
	void audio_unlock_transport ();

	int no_roll (jack_nframes_t nframes, jack_nframes_t offset, bool state_changing);
	
	atomic_t _state_changing;
	bool state_changing() const { return atomic_read (&_state_changing); }

	struct TransportRequest {
	    
	    enum Type {
		    SetTransportState,
		    StopAuditioning,
		    SendFullMTC,
		    SendMTC,
		    SendMMC,
		    EmitTransportStateChange
	    };

	    Type type;

	    State state;
	    jack_nframes_t locate_frame;

	    union {
		MIDI::MachineControl::Command mmc_cmd;
		bool abort;
	    };

	    TransportRequest () { 
		    abort = false;
	    }
	};

	PBD::Lock       midi_lock;
	pthread_t       transport_thread;
	PBD::Lock       transport_request_lock;
	int             transport_request_pipe[2];
	LockFreeFIFO<TransportRequest *> transport_requests;

	int           start_transport_thread ();
	void          terminate_transport_thread ();
	void          poke_transport_thread ();
	static void *_transport_thread_work (void *arg);
	void          transport_thread_work ();

	void          do_set_transport_state_in_another_thread (Session::State, 
								jack_nframes_t locate_frame = 0, 
								bool abort = false);
	void          set_transport_state (Session::State, 
								jack_nframes_t target_frame, bool abort);
	void          stop_transport (bool abort);
	void          locate (jack_nframes_t frame);

	TempoMap    *_tempo_map;
	void          tempo_map_changed();

	/* edit/mix groups */

	int load_route_groups (const XMLNode&, bool is_edit);
	int load_edit_groups (const XMLNode&);
	int load_mix_groups (const XMLNode&);


	list<RouteGroup *> edit_groups;
	list<RouteGroup *> mix_groups;

	/* disk-streams */

	DiskStreamList  diskstreams; 
	mutable PBD::Lock diskstream_lock;
	unsigned long dstream_buffer_size;
	void add_diskstream (DiskStream&);
	int  load_diskstreams (const XMLNode&);

	/* routes stuff */

	RouteList       routes;
	mutable PBD::NonBlockingLock route_lock;
	void   add_route (Route&);

	int load_routes (const XMLNode&);
	Route* XMLRouteFactory (const XMLNode&);

	/* mixer stuff */

	bool           _solo_latched;
	bool solo_update_disabled;
	unsigned long   current_block_size;

	void route_solo_changed (void *src, Route *);
	void catch_up_on_solo ();
	void update_route_solo_state ();
	void strip_portname_for_solo (string& portname);

	/* REGION MANAGEMENT */

	mutable PBD::Lock region_lock;
	typedef map<ARDOUR::id_t,Region *> RegionList;
	RegionList regions;
	
	void add_region (Region *);
	void remove_region (Region *);
	Region *get_region (ARDOUR::id_t);

	int load_regions (const XMLNode& node);

	/* SOURCES */
	
	PBD::Lock source_lock;
	typedef map<ARDOUR::id_t, Source *> SourceList;
	SourceList sources;

	int load_sources (const XMLNode& node);
	XMLNode& get_sources_as_xml ();

	void remove_source (Source *);

	Source *XMLSourceFactory (const XMLNode&);

	/* PLAYLISTS */
	
	mutable PBD::Lock playlist_lock;
	typedef set<Playlist *> PlaylistList;
	PlaylistList playlists;

	int load_playlists (const XMLNode&);
	void remove_playlist (Playlist *);

	Playlist *playlist_factory (string name);
	Playlist *XMLPlaylistFactory (const XMLNode&);

	void playlist_length_changed (Playlist *);
	void diskstream_playlist_changed (DiskStream *);

	/* NAMED SELECTIONS */

	mutable PBD::Lock named_selection_lock;
	typedef set<NamedSelection *> NamedSelectionList;
	NamedSelectionList named_selections;

	int load_named_selections (const XMLNode&);

	NamedSelection *named_selection_factory (string name);
	NamedSelection *XMLNamedSelectionFactory (const XMLNode&);

	/* DEFAULT FADE CURVES */

	float default_fade_steepness;
	float default_fade_msecs;

	/* AUDITIONING */

	Auditioner *auditioner;

	/* EXPORT */

	/* FLATTEN */

	int flatten_one_track (AudioTrack&, jack_nframes_t start, jack_nframes_t cnt);

	/* INSERT AND SEND MANAGEMENT */
	
	slist<PortInsert *>   _port_inserts;
	slist<PluginInsert *> _plugin_inserts;
	slist<Send *>         _sends;
	unsigned long          send_cnt;
	unsigned long          insert_cnt;

	void add_redirect (Redirect *);
	void remove_redirect (Redirect *);

	/* S/W RAID */

	struct space_and_path {
	    unsigned long blocks; /* 4kB blocks */
	    string path;
	    
	    space_and_path() { 
		    blocks = 0;
	    }
	};

	struct space_and_path_ascending_cmp {
	    bool operator() (space_and_path a, space_and_path b) {
		    return a.blocks > b.blocks;
	    }
	};
	
	string _raid_path;
	void setup_raid_path (string path);

	vector<space_and_path> session_dirs;
	vector<space_and_path>::iterator last_rr_session_dir;
	unsigned long _total_free_4k_blocks;
	PBD::Lock space_lock;

	string discover_best_sound_dir ();
	int ensure_sound_dir (string, string&);
	void refresh_disk_space ();

	/* I/O Connections */

	typedef vector<Connection *> ConnectionList;
	mutable PBD::Lock connection_lock;
	ConnectionList _connections;
	int load_connections (const XMLNode&);

	int set_slave_source_unlocked (SlaveSource, bool&);

	void reverse_diskstream_buffers ();

	UndoHistory history;
	UndoCommand current_cmd;

	GlobalRouteBooleanState get_global_route_boolean (bool (Route::*method)(void) const);
	void set_global_route_boolean (GlobalRouteBooleanState s, void (Route::*method)(bool, void*), void *arg);
	void set_global_mute (GlobalRouteBooleanState s, void *src);
	void set_global_solo (GlobalRouteBooleanState s, void *src);
	void set_global_record_enable (GlobalRouteBooleanState s, void *src);

	void set_jack_timebase ();
	void record_enable_change_all (bool yn);

	XMLNode& state(bool);

	/* click track */

	struct Click {
	    jack_nframes_t start;
	    jack_nframes_t duration;
	    jack_nframes_t offset;
	    const Sample *data;

	    Click (jack_nframes_t s, jack_nframes_t d, const Sample *b) 
		    : start (s), duration (d), data (b) { offset = 0; }
	    
	    static Pool* requestPool;

	    void *operator new(size_t ignored) {
		    return requestPool->alloc ();
	    };

	    void operator delete(void *ptr, size_t size) {
		    requestPool->release (ptr);
	    }
	};
 
	typedef list<Click*> Clicks;

	Clicks          clicks;
	Pool            click_pool;
	bool           _clicking;
	Port*          _click_port;
	Sample*         click_data;
	Sample*         click_emphasis_data;
	jack_nframes_t  click_length;
	jack_nframes_t  click_emphasis_length;

	static const Sample         default_click[];
	static const jack_nframes_t default_click_length;
	static const Sample         default_click_emphasis[];
	static const jack_nframes_t default_click_emphasis_length;

	Click *get_click();
	void   setup_click_sounds (int which);
	void   clear_clicks ();
	void   click (jack_nframes_t start, jack_nframes_t nframes, jack_nframes_t offset);
};

}; /* namespace ARDOUR */

#endif /* __ardour_session_h__ */
